# www.igvita.com

require 'rubygems'
require 'eventmachine'
require 'evma_httpserver'
require 'memcache'

class StaleWhileRevalidate  < EventMachine::Connection
  include EventMachine::HttpServer

  attr_accessor :memcached
  attr_reader :http_request_uri, :http_query_string
 
  def process_http_request
    resp = EventMachine::DelegatedHttpResponse.new(self)

    response = proc do |data|
      resp.status = 200
      resp.content = data
      resp.send_response
    end
 
    # Cache two keys: 'key' (with an expire) and 'key-no-ttl' (with no expire)
    # if 'key' has expired, return value at 'key-no-ttl', and start an async
    # update process to update the cache
    #   
    # key = full request path with query parameters
    cache_key = @http_request_uri.to_s + @http_query_string.to_s
    keys = {:key => cache_key, :no_ttl => cache_key + "-no-ttl", :processing => cache_key + "-processing"}
    cache = @memcached.get_multi(*keys.values)
     
    # async operation to compute value to be stored in the cache, in this case..
    # data is a simple timestamp. This is where actual app logic and  processing
    # is done. (talking to database, etc.)
    operation = proc do
      sleep 5
      @data = Time.now.to_s
    end
 
    # Callback to execute once the request is fulfilled, will update the
    # memcached values for future requests
    callback = proc do |res|
      
      # if cache is empty, then this is a new request, and we need to render
      # response back to user. This is the only time the user will hit live
      # application server
      response.call(@data) if cache.empty?
      
      @memcached.set(keys[:key], @data, 10) 
      @memcached.set(keys[:no_ttl], @data, 0) 
      @memcached.set(keys[:processing], 0, 0)
      
      puts "\t app server updated cache! New value for '#{keys[:key]}' : #{@data}"
    end

    # if key has not expired, then return immediately, we're safe!
    if cache[cache_key]
      puts "Valid cache for: #{cache_key} :: Data :: #{cache[cache_key]}"
      response.call(cache[cache_key])
      
    else
      # check if we've seen this request before... if so we'll have the no-ttl
      # key, which we'll return immediately, and then start an async process
      # to update the stale cache
      if cache[keys[:no_ttl]]
        puts "Stale cache for: #{keys[:key]} :: Data :: #{cache[keys[:no_ttl]]} :: Processing? :: #{cache[keys[:processing]]}"
        response.call(cache[keys[:no_ttl]])

        # set processing flag to true, this avoids multiple requests piling on
        # while the async query is waiting to complete
        @memcached.set(keys[:processing], 1, 0)
      end
    
      EM.defer(operation, callback) if cache[keys[:processing]].to_i != 1
    end
  end
end
 
EventMachine::run {
  @memcached = MemCache.new("localhost:11211") 
  EventMachine::start_server("0.0.0.0", 8081, StaleWhileRevalidate) { |conn| 
    conn.memcached = @memcached
  }
  puts "Listening on port 8081."
}
